/* avq_subs.c -- Subroutines for AVQ
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "avq_subs.h"
#include "stddefs.h"
#include "vector_utils.h"
#include "vector_read.h"
#include "cluster.h"

#define Abort(x)        AbortPrint((x), __LINE__, __FILE__);
#define SQR(x)          ((x) * (x))

void AbortPrint(int code, int line, char *file)
{
	fprintf(stderr, "Aborting... [%d | %s]\n", line, file);
	exit(code);
}


bool Allocate2DArray(char ***array, int xSize, int ySize, size_t elementSize)
{
        int     index;
 
        if ((*array = malloc(xSize * sizeof(void *))) == NULL)
                return FALSE;
 
        for (index = 0; index < xSize; index++) {   
                if (((*array)[index] = malloc(ySize * elementSize)) == NULL)
                        return FALSE;
         
                xSize--;
        }
}

codebook CodebookNew(int slots)
{
        codebook        newBook;
                
        if (!(newBook = (vector **) malloc(slots * sizeof(vector *))))
                return NULL;

        return newBook;
}

bool CheckClusterDimensions(cluster *clust)
{               
        int     dimensions;             /* dimension to check against */
        cluster *cIndex;                /* index into cluster */

        if (clust == NULL) return TRUE; /* empty cluster, so we pass */
 
        dimensions = clust -> vect -> dimensions;
        cIndex = clust -> next;         /* the first one passes by definition */
                
        while (cIndex != NULL) {
                if (cIndex -> vect -> dimensions != dimensions)
                        return FALSE;
                        
                cIndex = cIndex -> next;
        }
         
        return TRUE;
}

codebook CodebookNewFromCluster(cluster *clust, int *slots)
{
        codebook        newBook;                /* new codebook */
        cluster         *cIndex;                /* index into cluster */
        int             slot;                   /* slot index into codebook */

        if (!(newBook = CodebookNew(ClusterSize(clust))))
                return NULL;
                
        slot = 0;
        cIndex = clust;
        while (cIndex != NULL) {
                newBook[slot] = VectorMakeCopy(cIndex -> vect); 
                cIndex = cIndex -> next;
                slot++;
        }

	*slots = slot;
 
        return newBook;
}               

codebook CodebookDestroy(codebook book, int slots)
{
        int     slot;   
         
        slot = 0;
        while (slot < slots) {
                VectorDeallocate(book[slot]);
                slot++;
        }

        free(book);
        return NULL;
}


codebook CodebookMakeCopy(codebook book, int slots)
{
        codebook        newBook;
        int             slot;                   /* index into codebook slots */

        if (!(newBook = CodebookNew(slots)))
                return NULL;

        slot = 0;
        while (slot < slots) {
                newBook[slot] = VectorMakeCopy(book[slot]);
                slot++;
        }

        return newBook;
}

cluster **MakeClusters(int numSeeds)
{
        cluster **newClusterArray;
        int     slot;                   /* index into cluster array */

        if (!(newClusterArray = (cluster **) malloc(numSeeds * sizeof(cluster *)))) {
                fprintf(stderr, "*** Unable to allocate cluster space.\n");
                Abort(14);
        }

        slot = 0;
        while (slot < numSeeds) {
                newClusterArray[slot] = ClusterNew();
                slot++;
        }

        return newClusterArray;
}

cluster **DestroyClusters(cluster **clusters, int numSeeds)
{
        int     slot;

        slot = 0;
        while (slot < numSeeds) {
                clusters[slot] = ClusterDestroy(clusters[slot]);
                slot++;
        }

	free(clusters);

        return NULL;
}

int FindNearestOtherCluster(codebook currentBook, int numClusters, int target)
{
	int	slot, minSlot;
	double	minDist,
		dist;
	vector	*targetVect;

	minSlot = 0;
	targetVect = currentBook[target];
	slot = 0;

	if (target != 0) 
		minSlot = 0;
	else if (numClusters == 1)
		return 0;
	else
		minSlot = 1;

	minDist = VectorEuclideanDist(currentBook[target], currentBook[minSlot]);

	while (slot < numClusters) {
		if (slot != target) {
			if (currentBook[slot] != NULL) {
				dist = VectorEuclideanDist(targetVect, currentBook[slot]);
				if (dist < minDist) {
					minDist = dist;
					minSlot = slot;
				}
			}
		}
		slot++;
	}

	return minSlot;
}


int FindNearestNeighbour(vector *vect, codebook currentBook, int numClusters)
{
        int     slot;           /* index into codebook */
        double  minDist,        /* current minimum distance */
                dist;           /* current distance */
        int	minSlot;        /* current slot of minimum distance*/

        minSlot = 0;
        minDist = VectorEuclideanDist(vect, currentBook[0]);
        slot = 1;
        while (slot < numClusters) {
                if (currentBook[slot] != NULL) {
                        dist = VectorEuclideanDist(vect, currentBook[slot]);

                        if (dist < minDist) {
                                minDist = dist;
                                minSlot = slot;
                        }
                }
                slot++;
        }

        return minSlot;
}

double GetError(cluster **clusters, codebook book, int numClusters)
{
        double  bookError;
        int     slot;

        bookError = 0.0;
        slot = 0;
        while (slot < numClusters) {
                bookError += ClusterRMSDistance(clusters[slot], book[slot]);
                slot++;
        }

        bookError /= (double) numClusters;
        return sqrt(bookError);
}

double ClusterRMSDistance(cluster *clust, vector *vect)
{
        double  errSum;
        cluster *cIndex;

        if (ClusterIsEmpty(clust))  
                return 0.0;

        errSum = 0.0;
        cIndex = clust;
        while (cIndex != NULL) {
                errSum += SQR(VectorEuclideanDist(vect, cIndex -> vect));
                cIndex = cIndex -> next;
        }
                
        errSum /= (double) ClusterSize(clust);
        return sqrt(errSum);
}

double CodebookAverageDistance(vector *vect, codebook book, int numClusters)
{
        double  distance;
        int     slot;

        distance = 0.0;
        slot = 0;
        while (slot < numClusters) {
                if (book[slot] == NULL) {
                        numClusters--;
                        slot++;
                        continue;
                }
                distance += VectorEuclideanDist(vect, book[slot]);
                slot++;
        }
                
        return distance / (double) numClusters;
}

void WriteCodebook(FILE *output, codebook book, int slots)
{
        int     slot;

        for (slot = 0; slot < slots; slot++) {
                if (book[slot] != NULL) {
                        VectorWrite(output, book[slot]);
                        fprintf(output, "\n");
                }
        }
}               


codebook InitialiseCodebook(int numClusters, codebook vectorSet, int numSlots)
{
        codebook        newBook;
        int             slot;
        
        if (!(newBook = CodebookNew(numClusters))) {
                fprintf(stderr, "*** Could not initialise codebok.\n");
                Abort(20);
        }
        
        if (!(newBook[0] = CodebookCentroid(vectorSet, numSlots))) {
                fprintf(stderr, "*** Could not create initial centroid vector.\n");
                Abort(21);  
        }
        
        if (!(newBook[1] = VectorMakeCopy(vectorSet[0]))) {
                fprintf(stderr, "*** Could not allocate initial pattern vector.\n");
                Abort(23);
        }
                        
        for (slot = 2; slot < numClusters; slot++)
                newBook[slot] = NULL;
         
        return newBook;
}               

vector *CodebookCentroid(codebook book, int numSlots)
{
        vector  *vectSum;
        int     slot, entries;
 
        if (numSlots == 0)
                return NULL;

        if (!(vectSum = VectorZero(book[0] -> dimensions)))
                return NULL;

        entries = 0;
        slot = 0;
        while (slot < numSlots) {
                if (book[slot] != NULL) {
                        VectorSum(vectSum, book[slot]);
                        entries++;
                }
                slot++;
        }
        
        VectorDivideScalar(vectSum, (double) entries);
        return vectSum;     
}

int FreeCluster(codebook book, int numSlots)
{
        int     slot;

        for (slot = 0; slot < numSlots; slot++)
                if (book[slot] == NULL)
                        return slot;
                        
        return -1;
}               

cluster *ClusterFromCodebook(codebook book, int numSlots)
{
	cluster	*newClust;
	int	slot;

	newClust = NULL;
	slot = 0;

	while (slot < numSlots) {
		if (book[slot] != NULL)
			newClust = ClusterAddVector(newClust, VectorMakeCopy(book[slot]));
		slot++;
	}

	return newClust;
}

/* --- END of avq_subs.c --- */


