/* cluster : routines for dealing with clusters of vectors
 *
 * SEE cluster.h for details
 */

#include "cluster.h"
#include "vector_utils.h"
#include "vector_read.h"
#include "stddefs.h"
#include <stdlib.h>


cluster *ClusterAlloc(void)
{
	cluster	*clust;

	if ((clust = (cluster *) malloc(sizeof(cluster))) == NULL)
		return NULL;

	clust -> vect = NULL;
	clust -> next = NULL;

	return clust;
}


/* -- Cluster functions -- */

bool ClusterIsEmpty(cluster *clust)
{
	return (clust == NULL);
}

cluster *ClusterNew(void)
{
	return NULL;
}

cluster *ClusterDestroy(cluster *clust)
{
	if (ClusterIsEmpty(clust))
		return NULL;

	clust -> next = ClusterDestroy(clust -> next);
	VectorDeallocate(clust -> vect);
	free(clust);
	return NULL;
}

cluster *ClusterAddVector(cluster *clust, vector *vect)
{
	cluster	*newCluster;

	if (vect == NULL)
		return clust;

	if ((newCluster = ClusterAlloc()) == NULL) {
		ClusterDestroy(clust);
		return NULL;
	}

	newCluster -> vect = vect;
	newCluster -> next = clust;

	return newCluster;
}

cluster *ClusterMerge(cluster *clust1, cluster *clust2)
{
	cluster	*cIndex;

	if (ClusterIsEmpty(clust1))
		return clust2;

	if (ClusterIsEmpty(clust2))
		return clust1;

	cIndex = clust1;
	while (cIndex -> next != NULL)
		cIndex = cIndex -> next;

	cIndex -> next = clust2;
	return clust1;
}

vector *ClusterCentroid(cluster *clust)
{
	vector	*vectSum;
	cluster	*cIndex;		/* index into cluster */

	if (ClusterIsEmpty(clust))
		return NULL;

	if ((vectSum = VectorZero(clust -> vect -> dimensions)) == NULL)
		return NULL;

	cIndex = clust;
	while (cIndex != NULL) {
		VectorSum(vectSum, cIndex -> vect);
		cIndex = cIndex -> next;
	}

	VectorDivideScalar(vectSum, (double) ClusterSize(clust));

	return vectSum;
}

int ClusterSize(cluster *clust)
{
	int	size;
	cluster	*cIndex;

	if (ClusterIsEmpty(clust))
		return 0;

	size = 1;
	cIndex = clust;
	while (cIndex -> next != NULL) {
		size++;
		cIndex = cIndex -> next;
	}

	return size;
}

cluster *ClusterReadFromFile(FILE *inputFile)
{
	cluster	*clust = ClusterNew();
	bool		inFile;

	inFile = TRUE;
	while(inFile) {
		clust = ClusterAddVector(clust, VectorReadLine(inputFile, &inFile));
	}
	return clust;
}

cluster *ClusterRemoveVector(cluster *clust, vector *vect)
{
	cluster	*cIndex,
				*cTemp;

	if (vect == NULL)
		return clust;

	if (clust == NULL)
		return clust;

	if (VectorIdentical(clust -> vect, vect)) {
		cTemp = clust -> next;
		clust -> next = NULL;
		clust = ClusterDestroy(clust);
		return cTemp;
	}

	cIndex = clust;
	while (cIndex -> next != NULL) {
		if (VectorIdentical(cIndex -> next -> vect, vect)) {
			cTemp = cIndex -> next;
			cIndex -> next = cIndex -> next -> next;
			cTemp -> next = NULL;
			cTemp = ClusterDestroy(cTemp);
			return clust;
		}
		cIndex = cIndex -> next;
	}

	return clust;
}


bool ClusterVectorIsIn(cluster *clust, vector *vect)
{
	cluster	*cIndex;

	if (vect == NULL)
		return FALSE;

	if (clust == NULL)
		return FALSE;

	cIndex = clust;
	while (cIndex != NULL) {
		if (VectorIdentical(cIndex -> vect, vect)) {
			return TRUE;
		}
		cIndex = cIndex -> next;
	}

	return FALSE;
}


bool ClusterCheckDimensionality(cluster *clust)
{
	int		dimensions;             /* dimension to check against */
	cluster	*cIndex;                /* index into cluster */

	if (clust == NULL) return TRUE; /* empty cluster, so we pass */

	dimensions = clust -> vect -> dimensions;
	cIndex = clust -> next;         /* the first one passes by definition */

	while (cIndex != NULL) {
		if (cIndex -> vect -> dimensions != dimensions)
			return FALSE;

		cIndex = cIndex -> next;
	}

	return TRUE;
}


cluster *ClusterMakeCopy(cluster *clust)
{
	cluster	*newClust;

	if (clust == NULL)
		return NULL;

	newClust = NULL;
	newClust = ClusterAddVector(newClust, VectorMakeCopy(clust -> vect));
	newClust -> next = ClusterMakeCopy(clust -> next);

	return newClust;
}


/* --- END of cluster.c --- */


