/* k-means -- performs vector quantization on a set of vectors
 *            uses an adaptive modified Forgy's Algorithm to find 
 *            the means of a set of clusters that represent the
 *            entire set
 *
 * Author: Dylan Muir (dr.muir@student.qut.edu.au)
 *         QUT MLRC LPG Semester 2 1999
 * Date: 15th September, 1999
 * Modified: 19th October, 1999
 * Version: 0.10
 *
 * Usage: kmeans {input vectors file}  -- whitespace-delimited, one vector per line
 *                                        (accepts stdin if '-' is specified)
 *               {output means file}   -- tab-delimited, one mean vector per line
 *                                        (writes to stdout if '-' is specified)
 *               {av. dist. parameter} -- average distance a pattern must be from
 *                                        all clusters to create a new cluster
 *
 * Modules used: vector_utils, vector_read, cluster, stddefs, tokenlst, tokscan
 *
 * Aknowledgements: Towsey, M., "The Use of Neural Networks in the Automated Analysis of the Electroencephalogram"
 *                  Sayood, K., "Introduction to Data Compression"
 *                  Das, S. and Mozer, M., "Dynamic On-Line Clustering and State Extraction: An Approach to Symbolic Learning"
 */

/* -- Required modules -- */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "vector_utils.h"
#include "vector_read.h"
#include "cluster.h"
#include "stddefs.h"


/* -- kmeans defines -- */

#define	APP_NAME	0
#define	VECTOR_FILE	1
#define HELP		1
#define	MEANS_FILE	2
#define	DIST_PARM	3
#define	NUMARGS		4

#define	VERSION	"0.10"

#define	MAX_CLUSTERS	200

#define	Abort(x)	AbortPrint((x), __LINE__, __FILE__);
#define	SQR(x)		((x) * (x))
#define	ABS(x)		((x) < 0 ? (-(x)) : (x))

/* -- kmeans typedefs -- */

typedef	vector	**codebook;		/* an array of vector pointers */


/* -- kmeans helper function prototypes -- */

void Usage(FILE *output, char *appName);
void Help(FILE *output, char *appName);
void AbortPrint(int code, int line, char *file);
void ProcessCommandLine(int argc, char *argv[],
			FILE **vectorFile, FILE **meansFile, float *distParm);
void CloseFiles(FILE *vectorFile, FILE *meansFile);
void PrintBanner(char *argv[], FILE *vectorFile, FILE *meansFile, float distParm);
bool Allocate2DArray(char ***array, int xSize, int ySize, size_t elementSize);
bool CheckClusterDimensions(cluster *clust);
codebook ReadTrainingSet(FILE *vectorFile, int *numVectors);
cluster **MakeClusters(int numClusters);
cluster **DestroyClusters(cluster **clusters, int numClusters);
int FindNearestNeighbour(vector *vect, codebook currentBook, int numClusters);
double GetError(cluster **clusters, codebook book, int numClusters);
double ClusterRMSDistance(cluster *clust, vector *vect);
codebook InitialiseCodebook(int numClusters, codebook vectorSet, int numSlots);
double CodebookAverageDistance(vector *vect, codebook book, int numClusters);
vector *CodebookCentroid(codebook book, int numSlots);
int FreeCluster(codebook book, int numSlots);


/* -- Codebook functions -- */
codebook CodebookNew(int slots);
codebook CodebookDestroy(codebook book, int slots);
codebook CodebookNewFromCluster(cluster *clust);
codebook CodebookMakeCopy(codebook book, int slots);
void WriteCodebook(FILE *output, codebook book, int slots);


/* -- main -- */

int main(int argc, char *argv[])
{
	FILE	*vectorFile, *meansFile;
	int	lastCluster;
	bool	randomSeeds,
		terminate;
	float	distParm;
	double	lastError, thisError;

	codebook	seeds,		/* The initial seeds */
			currentBook,	/* The current set of centroids */
			nextBook,	/* The next codebook */
			vectorSet;	/* Set of all training vectors */

	int	slot,
		oldClosest,
		numTrainVectors,	/* How many vectors read from traing file */
		*clusterMembership;	/* What cluster does this vector belong to? */

	cluster	**clusters;		/* Array of current clusters */

	cluster	*cIndex;
	double	avDist, closestDist;
	int	closest;


	srand48(((double) time(0)) + ((double) getpid()));	/* Initiallise rand48 generator */

	ProcessCommandLine(	argc, argv,
				&vectorFile, &meansFile, &distParm);

	PrintBanner(argv, vectorFile, meansFile, distParm);
	vectorSet = ReadTrainingSet(vectorFile, &numTrainVectors);

	if (!(clusterMembership = (int *) malloc(numTrainVectors * sizeof(int)))) {
		fprintf(stderr, "*** Could not allocate cluster emmbership array.\n");
		Abort(24);
	}

	for (slot = 0; slot < numTrainVectors; slot++)
		clusterMembership[slot] = -1;

	currentBook = InitialiseCodebook(MAX_CLUSTERS, vectorSet, numTrainVectors);
	lastCluster = 2;
	clusters = MakeClusters(MAX_CLUSTERS);
	lastError = 999.0;

	terminate = FALSE;
	while (!terminate) {
		slot = 0;
		while (slot < numTrainVectors) {
			avDist = CodebookAverageDistance(vectorSet[slot], currentBook, lastCluster);
			closest = FindNearestNeighbour(vectorSet[slot], currentBook, lastCluster);
			closestDist = VectorEuclideanDist(vectorSet[slot], currentBook[closest]);

/*			if (ABS(closestDist - avDist) > (avDist * distParm)) { */
			if (closestDist > (avDist * distParm)) {
				closest = FreeCluster(currentBook, MAX_CLUSTERS);
				if (closest == -1) {
					fprintf(stderr, "*** Exceeded compiler limit of clusters [%d].\n", MAX_CLUSTERS);
					Abort(22);
				}
				if (closest > lastCluster)
					lastCluster = closest;
			}

			oldClosest = clusterMembership[slot];

			if (oldClosest != -1) {
				clusters[oldClosest] = ClusterRemoveVector(clusters[oldClosest], vectorSet[slot]);
				if (currentBook[oldClosest] != NULL) VectorDeallocate(currentBook[oldClosest]);
				currentBook[oldClosest] = ClusterCentroid(clusters[oldClosest]);
			}

			clusters[closest] = ClusterAddVector(clusters[closest], VectorMakeCopy(vectorSet[slot]));
			clusterMembership[slot] = closest;

			if (currentBook[closest] != NULL) VectorDeallocate(currentBook[closest]);
			currentBook[closest] = ClusterCentroid(clusters[closest]);

			slot++;
		}

		thisError = GetError(clusters, currentBook, lastCluster);
		if ((lastError >= thisError) && (lastError - thisError < 0.0001))
			terminate = TRUE;

		fprintf(stderr, "RMS error: %f (%d clusters)\n", thisError, lastCluster);
		lastError = thisError;
	}

	WriteCodebook(meansFile, currentBook, MAX_CLUSTERS);

	CloseFiles(vectorFile, meansFile);

	return 0;
}


/* -- helper functions -- */

void Usage(FILE *output, char *appName)
{
	fprintf(output, "\n*** k-means -- Performs adaptive clustering on a set of vectors\n");
	fprintf(output, "    Version %s build %s %s\n", VERSION, __TIME__, __DATE__);
	fprintf(output, "Usage: %s {vector file} {means file} {av. dist. parm}\n", appName);
	fprintf(output, "Where: {vector file}    -- file of vectors to analyse, one per line\n");
	fprintf(output, "                           ['-' for stdin]\n");
	fprintf(output, "       {means file}     -- filename to output the mean vectors to\n");
	fprintf(output, "                           ['-' for stdout]\n");
	fprintf(output, "       {av. dist. parm} -- see help file for information\n");
	fprintf(output, "                           (0.5 is a good value)\n");
	fprintf(output, "For detailed info, type %s --help\n", appName);
	fprintf(output, "\n");
}

void AbortPrint(int code, int line, char *file)
{
	fprintf(stderr, "Aborting [%s | %d]...\n", file, line);
	exit(code);
}


void ProcessCommandLine(int argc, char *argv[], FILE **vectorFile, FILE **meansFile, float *distParm)
{
	if ((argc > 1) && (strcasecmp(argv[HELP], "--help") == 0)) {
		FILE	*helpFile;

		if (!(helpFile = fopen("kmeans.help", "wt"))) {
			fprintf(stderr, "*** Could not create help file [kmeans.help].\n");
			Abort(15);
		}

		Help(helpFile, argv[APP_NAME]);
		printf("Created kmeans.help\n");
		exit(0);
	}

	if (argc < NUMARGS) {
		Usage(stderr, argv[APP_NAME]);
		fprintf(stderr, "*** Invalid number of arguments.\n");
		Abort(1);
	}

	if (argc > NUMARGS)
		fprintf(stderr, "--- Warning: extra arguments ignored.\n");

	if (strcmp(argv[VECTOR_FILE], "-") == 0)
		*vectorFile = stdin;

	else if (!(*vectorFile = fopen(argv[VECTOR_FILE], "rt"))) {
		Usage(stderr, argv[APP_NAME]);
		fprintf(stderr, "*** Could not open vector file [%s] for reading.\n", argv[VECTOR_FILE]);
		Abort(2);
	}

	if (strcmp(argv[MEANS_FILE], "-") == 0)
		*meansFile = stdout;

	else if (!(*meansFile = fopen(argv[MEANS_FILE], "wt"))) {
		Usage(stderr, argv[APP_NAME]);
		fprintf(stderr, "*** Could not open means file [%s] for writing.\n", argv[MEANS_FILE]);
		Abort(3);
	}

	if (!isdigit(*argv[DIST_PARM])) {
		Usage(stderr, argv[APP_NAME]);
		fprintf(stderr, "*** Error converting distance parameter from [%s].\n", argv[DIST_PARM]);
		Abort(5);
	}

	*distParm = atof(argv[DIST_PARM]);
	if ((*distParm > 1.0) || (*distParm < 0.0)) {
		fprintf(stderr, "*** Distance parameter [%2f] must be between 0.0 and 1.0.\n");
		Abort(26);
	}
}

void CloseFiles(FILE *vectorFile, FILE *meansFile)
{
	if (vectorFile != stdin) fclose(vectorFile);
	if (meansFile != stdout) fclose(meansFile);
}


void PrintBanner(char *argv[], FILE *vectorFile, FILE *meansFile, float distParm)
{
	fprintf(stderr, "- k-means ver %s\n", VERSION);
	fprintf(stderr, " o Vectors: << [");
	vectorFile == stdin ?
		fprintf(stderr, "- stdin]\n") :
		fprintf(stderr, "%s]\n", argv[VECTOR_FILE]);

	fprintf(stderr, " o Means: >> [");
	meansFile == stdout ?
		fprintf(stderr, "- stdout]\n") :
		fprintf(stderr, "%s]\n", argv[MEANS_FILE]);

	fprintf(stderr, " o distance parameter: [%f]\n", distParm);
}
	

bool Allocate2DArray(char ***array, int xSize, int ySize, size_t elementSize)
{
        int     index;
 
        if ((*array = malloc(xSize * sizeof(void *))) == NULL)
                return FALSE;
 
        for (index = 0; index < xSize; index++) {
                if (((*array)[index] = malloc(ySize * elementSize)) == NULL)
                        return FALSE;
 
                xSize--;
        }
}

codebook CodebookNew(int slots)
{
	codebook	newBook;

        if (!(newBook = (vector **) malloc(slots * sizeof(vector *))))
		return NULL;

	return newBook;
}

bool CheckClusterDimensions(cluster *clust)
{
	int	dimensions;		/* dimension to check against */
	cluster	*cIndex;		/* index into cluster */

	if (clust == NULL) return TRUE;	/* empty cluster, so we pass */

	dimensions = clust -> vect -> dimensions;
	cIndex = clust -> next;		/* the first one passes by definition */

	while (cIndex != NULL) {
		if (cIndex -> vect -> dimensions != dimensions)
			return FALSE;

		cIndex = cIndex -> next;
	}

	return TRUE;
}

codebook CodebookNewFromCluster(cluster *clust)
{
	codebook	newBook;		/* new codebook */
	cluster		*cIndex;		/* index into cluster */
	int		slot;			/* slot index into codebook */

	if (!(newBook = CodebookNew(ClusterSize(clust))))
		return NULL;

	slot = 0;
	cIndex = clust;
	while (cIndex != NULL) {
		newBook[slot] = VectorMakeCopy(cIndex -> vect);
		cIndex = cIndex -> next;
		slot++;
	}

	return newBook;
}

codebook CodebookDestroy(codebook book, int slots)
{
	int	slot;

	slot = 0;
	while (slot < slots) {
		VectorDeallocate(book[slot]);
		slot++;
	}

	free(book);
	return NULL;
}

codebook ReadTrainingSet(FILE *vectorFile, int *numVectors)
{
	cluster		*vectorSet;
	codebook	trainingSet;

	fprintf(stderr, "Reading training vectors...\n");
	vectorSet = ClusterReadFromFile(vectorFile);
	if (ClusterIsEmpty(vectorSet)) {
		fprintf(stderr, "*** Empty training set.\n");
		Abort(10);
	}

	*numVectors = ClusterSize(vectorSet);
	fprintf(stderr, "Read %d vectors.\n", *numVectors);

	if (!CheckClusterDimensions(vectorSet)) {	/* make sure they're all the same dimensionality */
		fprintf(stderr, "*** The training vectors must all be the same dimensionality.\n");
		Abort(7);
	}
	fprintf(stderr, "Training dimension: %d\n", vectorSet -> vect -> dimensions);

	trainingSet =  CodebookNewFromCluster(vectorSet);
	vectorSet = ClusterDestroy(vectorSet);
	return trainingSet;
}

codebook CodebookMakeCopy(codebook book, int slots)
{
	codebook	newBook;
	int		slot;			/* index into codebook slots */

	if (!(newBook = CodebookNew(slots)))
		return NULL;

	slot = 0;
	while (slot < slots) {
		newBook[slot] = VectorMakeCopy(book[slot]);
		slot++;
	}

	return newBook;
}

cluster **MakeClusters(int numSeeds)
{
	cluster	**newClusterArray;
	int	slot;			/* index into cluster array */

	if (!(newClusterArray = (cluster **) malloc(numSeeds * sizeof(cluster *)))) {
		fprintf(stderr, "*** Unable to allocate cluster space.\n");
		Abort(14);
	}

	slot = 0;
	while (slot < numSeeds) {
		newClusterArray[slot] = ClusterNew();
		slot++;
	}

	return newClusterArray;
}
	
cluster **DestroyClusters(cluster **clusters, int numSeeds)
{
	int	slot;

	slot = 0;
	while (slot < numSeeds) {
		clusters[slot] = ClusterDestroy(clusters[slot]);
		slot++;
	}

	return NULL;
}


int FindNearestNeighbour(vector *vect, codebook currentBook, int numClusters)
{
	int	slot;		/* index into codebook */
	double	minDist,	/* current minimum distance */
		dist;		/* current distance */
	double	minSlot;	/* current slot of minimum distance*/

	minSlot = 0;
	minDist = VectorEuclideanDist(vect, currentBook[0]);
	slot = 1;
	while (slot < numClusters) {
		if (currentBook[slot] != NULL) {
			dist = VectorEuclideanDist(vect, currentBook[slot]);

			if (dist < minDist) {
				minDist = dist;
				minSlot = slot;
			}
		}
		slot++;
	}

	return minSlot;
}

double GetError(cluster **clusters, codebook book, int numClusters)
{
	double	bookError;
	int	slot;

	bookError = 0.0;
	slot = 0;
	while (slot < numClusters) {
		bookError += ClusterRMSDistance(clusters[slot], book[slot]);
		slot++;
	}

	bookError /= (double) numClusters;
	return sqrt(bookError);
}

double ClusterRMSDistance(cluster *clust, vector *vect)
{
	double	errSum;
	cluster	*cIndex;

	if (ClusterIsEmpty(clust))
		return 0.0;

	errSum = 0.0;
	cIndex = clust;
	while (cIndex != NULL) {
		errSum += SQR(VectorEuclideanDist(vect, cIndex -> vect));
		cIndex = cIndex -> next;
	}

	errSum /= (double) ClusterSize(clust);
	return sqrt(errSum);
}

double CodebookAverageDistance(vector *vect, codebook book, int numClusters)
{
	double	distance;
	int	slot;

	distance = 0.0;
	slot = 0;
	while (slot < numClusters) {
		if (book[slot] == NULL) {
			numClusters--;
			slot++;
			continue;
		}
		distance += VectorEuclideanDist(vect, book[slot]);
		slot++;
	}

	return distance / (double) numClusters;
}


void WriteCodebook(FILE *output, codebook book, int slots)
{
	int	slot;

	for (slot = 0; slot < slots; slot++) {
		if (book[slot] != NULL) {
			VectorWrite(output, book[slot]);
			fprintf(output, "\n");
		}
	}
}

void Help(FILE *helpFile, char *appName)
{
	fprintf(helpFile, "-------------------------------------\n");
	fprintf(helpFile, "   Help file for K-means ver %s\n", VERSION);
	fprintf(helpFile, "-------------------------------------\n\n");

	Usage(helpFile, appName);

#include "kmeans.help.make"
}

codebook InitialiseCodebook(int numClusters, codebook vectorSet, int numSlots)
{
	codebook	newBook;
	int		slot;

	if (!(newBook = CodebookNew(numClusters))) {
		fprintf(stderr, "*** Could not initialise codebok.\n");
		Abort(20);
	}

	if (!(newBook[0] = CodebookCentroid(vectorSet, numSlots))) {
		fprintf(stderr, "*** Could not create initial centroid vector.\n");
		Abort(21);
	}

	if (!(newBook[1] = VectorMakeCopy(vectorSet[0]))) {
		fprintf(stderr, "*** Could not allocate initial pattern vector.\n");
		Abort(23);
	}

	for (slot = 2; slot < numClusters; slot++)
		newBook[slot] = NULL;

	return newBook;
}

vector *CodebookCentroid(codebook book, int numSlots)
{
	vector	*vectSum;
	int	slot, entries;

	if (numSlots == 0)
		return NULL;

	if (!(vectSum = VectorZero(book[0] -> dimensions)))
		return NULL;

	entries = 0;
	slot = 0;
	while (slot < numSlots) {
		if (book[slot] != NULL) {
			VectorSum(vectSum, book[slot]);
			entries++;
		}
		slot++;
	}

	VectorDivideScalar(vectSum, (double) entries);
	return vectSum;
}

int FreeCluster(codebook book, int numSlots)
{
	int	slot;

	for (slot = 0; slot < numSlots; slot++)
		if (book[slot] == NULL)
			return slot;

	return -1;
}


/* --- END of kmeans.c --- */


