/* tlavq.c - simulator for arbitrary networks with time-ordered input */
/*           Includes vector quantisation -- Dylan Muir October 1999  */

/*------------------------------------------------------------------------

This program simulates learning in a neural network using either
the classical back-propagation learning algorithm or a slightly
modified form derived in Williams and Zipser, "A Learning Algo-
rithm for Continually Running Fully Recurrent Networks."  The
input is a sequence of vectors of (ascii) floating point numbers
contained in a ".data" file.  The target outputs are a set of
time-stamped vectors of (ascii) floating point numbers (including
optional "don't care" values) in a ".teach" file.  The network
configuration is defined in a ".cf" file documented in tlearn.man.

------------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <signal.h>
#ifdef ibmpc
#include "strings.h"
#include <fcntl.h>
#else
#include <string.h>
#include <sys/file.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>

#include "vector_utils.h"
#include "cluster.h"
#include "avq.h"

#ifdef ibmpc
#define random(x) rand(x)
#define srandom(x) srand(x)
#endif

int	nn;		/* number of nodes */
int	ni;		/* number of inputs */
int	no;		/* number of outputs */
int	nt;		/* nn + ni + 1 */
int	np;		/* ni + 1 */
int	nhc;		/* number of hidden/context units */

struct	cf {
	int	con;	/* connection flag */
	int	fix;	/* fixed-weight flag */
	int	num;	/* group number */
	int	lim;	/* weight-limits flag */
	float	min;	/* weight minimum */
	float	max;	/* weight maximum */
};

struct	nf {
	int	func;	/* activation function type */
	int	dela;	/* delay flag */
	int	targ;	/* target flag */
};

struct	cf	**cinfo;	/* (nn x nt) connection info */
struct	nf	*ninfo;		/* (nn) node activation function info */

int	*outputs;	/* (no) indices of output nodes */
int	*selects;	/* (nn+1) nodes selected for probe printout */
int	*linput;	/* (ni) localist input array */
int	*clustnodes;	/* (nn+1) nodes to use avq on */
int	*contxtnodes;	/* (nn+1) nodes defined as context units */

long	epochlength = 0;	/* number of sweeps in one epoch */
long	epochsforavq = 2;	/* number of epochs to gather for avq */
long	currentepoch = 0;	/* current epoch we're in */
int	codebook_size;		/* size of codebook (number of clusters */
vector	**avq_codebook;		/* current AVQ codebook */
vector	**patterns;		/* HU patterns */
long	patterns_index = 0;	/* index into patterns array */
long	num_patterns;		/* max number of patterns */

double	avq_phi = 60;	/* 'phi' parameter for avq */
double	avq_t = 0.5;	/* 'T' parameter for avq */

float	*znew;		/* (nt) inputs and activations at time t+1 */
float	*zold;		/* (nt) inputs and activations at time t */
float	*zmem;		/* (nt) inputs and activations at time t */
float	**wt;		/* (nn x nt) weight TO node i FROM node j*/ 
float	**dwt;		/* (nn x nt) delta weight at time t */
float	**winc;		/* (nn x nt) accumulated weight increment*/
float	*target;	/* (no) output target values */
float	*error;		/* (nn) error = (output - target) values */
float	***pnew;	/* (nn x nt x nn) p-variable at time t+1 */
float	***pold;	/* (nn x nt x nn) p-variable at time t */

float	rate = .1;	/* learning rate */
float	momentum = 0.;	/* momentum */
float	weight_limit = 1.;	/* bound for random weight init */
float	criterion = 0.;	/* exit program when rms error is less than this */
float	init_bias = 0.;	/* possible offset for initial output biases */

float	err = 0.;	/* cumulative ss error */
float	ce_err = 0.;	/* cumulate cross_entropy error */
float	avq_err;	/* error for avq modification */

long	sweep = 0;	/* current sweep */
long	err_sweeps = 0;	/* number of sweeps for cumulative error */
long	tsweeps = 0;	/* total sweeps to date */
long	report = 0;	/* output rms error every "report" sweeps */

int	ngroups = 0;	/* number of groups */

int	backprop = 1;	/* flag for standard back propagation (the default) */
int	teacher = 0;	/* flag for feeding back targets */
int	localist = 0;	/* flag for speed-up with localist inputs */
int	randomly = 0;	/* flag for presenting inputs in random order */
int	limits = 0;	/* flag for limited weights */
int	ce = 0;		/* flag for cross_entropy */
int	avq_flag = 0;	/* flag to perform avq */
int	loadflag = 0;	/* flag for loading initial weights from file */

char	root[128];	/* root filename for .cf, .data, .teach, etc.*/
char	loadfile[128];	/* filename for weightfile to be read in */

FILE	*cfp;		/* file pointer for .cf file */

void intr();

extern int load_wts();
extern int save_wts();
extern int act_nds();


main(argc,argv)
	int	argc;
	char  **argv;
{

	FILE	*fopen();
	extern	char *optarg;
	extern	float rans();
	extern	long time();
/*	extern	int intr(); */


	long	nsweeps = 0;	/* number of sweeps to run for */
	long	ttime = 0;	/* number of sweeps since time = 0 */
	long	utime = 0;	/* number of sweeps since last update_weights */
	long	tmax = 0;	/* maximum number of sweeps (given in .data) */
	long	umax = 0;	/* update weights every umax sweeps */
	long	rtime = 0;	/* number of sweeps since last report */
	long	check = 0;	/* output weights every "check" sweeps */
	long	ctime = 0;	/* number of sweeps since last check */

	int	c;
	int	i;
	int	j;
	int	k;
	int	nticks = 1;	/* number of internal clock ticks per input */
	int	ticks = 0;	/* counter for ticks */
	int	learning = 1;	/* flag for learning */
	int	reset = 0;	/* flag for resetting net */
	int	verify = 0;	/* flag for printing output values */
	int	probe = 0;	/* flag for printing selected node values */
	int	command = 1;	/* flag for writing to .cmd file */
	int	iflag = 0;	/* flag for -I */
	int	tflag = 0;	/* flag for -T */
	int	rflag = 0;	/* flag for -x */
	int	seed = 0;	/* seed for random() */
	float	avq_err_bias = 0.0;	/* offset for alpha calculation */

	float	*w;
	float	*wi;
	float	*dw;
	float	*pn;
	float	*po;

	struct	cf	*ci;

	char	cmdfile[128];	/* filename for logging runs of program */
	char	cfile[128];	/* filename for .cf file */

	FILE	*cmdfp;

	signal(SIGINT, intr);
#ifndef ibmpc
	signal(SIGHUP, intr);
	signal(SIGQUIT, intr);
	signal(SIGKILL, intr);
#endif

#ifndef ibmpc
	exp_init();
#endif

#ifdef	DEBUG
	fprintf(stderr,"TlearnAVQ build [%s %s] DEBUG\n", __DATE__, __TIME__);
#endif

	root[0] = 0;
	nhc = 0;

	while ((c = getopt(argc, argv, "o:f:hiql:m:n:r:s:e:d:p:tC:E:ILM:PRS:TU:VXB:H:")) != EOF) {
		switch (c) {
			case 'o':
				avq_err_bias = atof(optarg);
#ifdef	DEBUG
				fprintf(stderr, "AVQ RMS error bias: %f\n", avq_err_bias);
#endif
				break;
			case 'q':
				avq_flag = 1;
#ifdef	DEBUG
				fprintf(stderr, "Performing AVQ on hidden units\n");
#endif
				break;
			case 'e':
				epochlength = (long) atol(optarg);
#ifdef	DEBUG
				fprintf(stderr, "Sweeps per epoch: %ld\n", epochlength);
#endif
				break;
			case 'p':
				avq_phi = (double) atof(optarg);
#ifdef	DEBUG
				fprintf(stderr, "AVQ Phi parameter: %f\n", avq_phi);
#endif
				break;
			case 'd':
				avq_t = (double) atof(optarg);
#ifdef	DEBUG
				fprintf(stderr, "AVQ T parameter: %f\n", avq_t);
#endif
				break;
			case 'C':
				check = (long) atol(optarg);
				ctime = check;
#ifdef	DEBUG
				fprintf(stderr, "Weights checkpoint: %ld\n", check);
#endif
				break;
			case 'f':
				strcpy(root,optarg);
#ifdef	DEBUG
				fprintf(stderr, "Input file base: [%s]\n", optarg);
#endif
				break;
			case 'i':
				command = 0;
				break;
			case 'l':
				loadflag = 1;
				strcpy(loadfile,optarg);
#ifdef	DEBUG
				fprintf(stderr, "Load state file base: [%s].wts(, .means)\n", optarg);
#endif
				break;
			case 'm':
				momentum = (float) atof(optarg);
#ifdef	DEBUG
				fprintf(stderr, "Momentum: %f\n", momentum);
#endif
				break;
			case 'n':
				nticks = (int) atoi(optarg);
#ifdef	DEBUG
				fprintf(stderr, "Clock ticks per input: %d\n", nticks);
#endif
				break;
			case 'P':
				probe = 1;
				learning = 0;
#ifdef	DEBUG
				fprintf(stderr, "Probe mode on\n");
#endif
				break;
			case 'r':
				rate = (float) atof(optarg);
#ifdef	DEBUG
				fprintf(stderr, "Learning rate: %f\n", rate);
#endif
				break;
			case 's':
				nsweeps = (long) atol(optarg);
#ifdef	DEBUG
				fprintf(stderr, "Sweeps to execute: %ld\n", nsweeps);
#endif
				break;
			case 't':
				teacher = 1;
#ifdef	DEBUG
				fprintf(stderr, "Feedback teacher values mode on\n");
#endif
				break;
			case 'L':
				backprop = 0;
#ifdef	DEBUG
				fprintf(stderr, "RTRL mode on\n");
#endif
				break;
			case 'V':
				verify = 1;
				learning = 0;
#ifdef	DEBUG
				fprintf(stderr, "Verify mode on\n");
#endif
				break;
			case 'X':
				rflag = 1;
#ifdef	DEBUG
				fprintf(stderr, "Using .reset file\n");
#endif
				break;
			case 'E':
				report = (long) atol(optarg);
#ifdef	DEBUG
				fprintf(stderr, "RMS checkpoint: %ld\n", report);
#endif
				break;
			case 'I':
				iflag = 1;
#ifdef	DEBUG
				fprintf(stderr, "Ignoring input values during extra clock ticks\n");
#endif
				break;
			case 'M':
				criterion = (float) atof(optarg);
#ifdef	DEBUG
				fprintf(stderr, "RMS target value for termination: %f\n", criterion);
#endif
				break;
			case 'R':
				randomly = 1;
#ifdef	DEBUG
				fprintf(stderr, "Randomly presenting input patterns\n");
#endif
				break;
			case 'S':
				seed = atoi(optarg);
#ifdef	DEBUG
				fprintf(stderr, "RNG seed: %d\n", seed);
#endif
				break;
			case 'T':
				tflag = 1;
#ifdef	DEBUG
				fprintf(stderr, "Ignoring target values during extra clock ticks\n");
#endif
				break;
			case 'U':
				umax = atol(optarg);
				break;
			case 'B':
				init_bias = atof(optarg);
#ifdef	DEBUG
				fprintf(stderr, "Initial bias offset: %f\n", init_bias);
#endif
				break;
			/*
			 * if == 1, use cross-entropy as error;
			 * if == 2, also collect cross-entropy stats.
			 */
			case 'H':
				ce = atoi(optarg);
				break;
			case '?':
			case 'h':
			default:
				usage();
				exit(2);
			break;
		}
	}

	if (nsweeps == 0){
		perror("ERROR: No -s specified");
		exit(1);
	}

	if (avq_flag && (epochlength == 0)) {
		perror("ERROR: No -e specified for AVQ mode");
		exit(1);
	}

	/* open files */

	if (root[0] == 0){
		perror("ERROR: No fileroot specified");
		exit(1);
	}

	if (command) {
		sprintf(cmdfile, "%s.cmd", root);
		cmdfp = fopen(cmdfile, "a");
		if (cmdfp == NULL) {
			perror("ERROR: Can't open .cmd file");
			exit(1);
		}
		for (i = 1; i < argc; i++)
			fprintf(cmdfp,"%s ",argv[i]);
		fprintf(cmdfp,"\n");
		fflush(cmdfp);
	}

	sprintf(cfile, "%s.cf", root);
	cfp = fopen(cfile, "r");
	if (cfp == NULL) {
		perror("ERROR: Can't open .cf file");
		exit(1);
	}

	get_nodes();
	make_arrays();
	get_outputs();
	get_connections();
	get_special();

	if (nhc == 0) {
		perror("ERROR: Hidden and context nodes must be specified");
		exit(1);
	}

	if (avq_flag) AVQ_check_hidden_context();

	if (!seed)
		time(&seed);
	srandom(seed);

	if (loadflag)
		load_wts();
	else {
		for (i = 0; i < nn; i++){
			w = *(wt + i);
			dw = *(dwt+ i);
			wi = *(winc+ i);
			ci = *(cinfo+ i);
			for (j = 0; j < nt; j++, ci++, w++, wi++, dw++){
				if (ci->con) 
					*w = rans(weight_limit);
				else
					*w = 0.;
				*wi = 0.;
				*dw = 0.;
			}
		}
		/*
		 * If init_bias, then we want to set initial biases
		 * to (*only*) output units to a random negative number.
		 * We index into the **wt to find the section of receiver
		 * weights for each output node.  The first weight in each
		 * section is for unit 0 (bias), so no further indexing needed.
		 */
		for (i = 0; i < no; i++){
			w = *(wt + outputs[i] - 1);
			ci = *(cinfo + outputs[i] - 1); 
			if (ci->con) 
				*w = init_bias + rans(.1);
			else
				*w = 0.;
		}
	}
	zold[0] = znew[0] = 1.;
	for (i = 1; i < nt; i++)
		zold[i] = znew[i] = 0.;
	if (backprop == 0){
		make_parrays();
		for (i = 0; i < nn; i++){
			for (j = 0; j < nt; j++){
				po = *(*(pold + i) + j);
				pn = *(*(pnew + i) + j);
				for (k = 0; k < nn; k++, po++, pn++){
					*po = 0.;
					*pn = 0.;
				}
			}
		}
	}


	nsweeps += tsweeps;
	sweep = tsweeps;
	err_sweeps = 0;
	currentepoch = 0;
	while (sweep < nsweeps) {

/*	for (sweep = tsweeps; sweep < nsweeps; sweep++){*/
		for (ticks = 0; ticks < nticks; ticks++){

			update_reset(ttime,ticks,rflag,&tmax,&reset);

			if (reset){
				if (backprop == 0)
					reset_network(zold,znew,pold,pnew);
				else
					reset_bp_net(zold,znew);
			}

			update_inputs(zold,ticks,iflag,&tmax,&linput);

			if (avq_flag && (sweep == tsweeps)) {		/* need to initialise AVQ */
				num_patterns = tmax * epochsforavq;
				AVQ_initialise();
			}

			if (learning || teacher || (report != 0))
				update_targets(target,ttime,ticks,tflag,&tmax);

			act_nds(zold,zmem,znew,wt,linput,target);

			if (learning) {	/* modified to always calculate error */
				comp_errors(zold,target,error,&err,&ce_err);
				avq_err = sqrt(err / ((double) err_sweeps + 1.0) / (double) no) - avq_err_bias;
			}	/* Note: by here, we've really done another sweep */
				/* avq_err = sqrt(mean_err) / sqrt(num outputs) */

			if (avq_flag) {
				if (patterns_index < num_patterns) AVQ_add_pattern(zold);
				AVQ_classify_hu_to_context(zold, zmem, znew, wt, linput, target);
			}

			if (learning && (backprop == 0))
				comp_deltas(pold,pnew,wt,dwt,zold,znew,error);
			if (learning && (backprop == 1))
			   comp_backprop(wt,dwt,zold,zmem,target,error,linput);

			if (probe)
				print_nodes(zold);
		}

		if (avq_flag && (sweep % epochlength == 0) && (sweep > tsweeps)) {
			currentepoch++;
			if (currentepoch >= epochsforavq) {
				AVQ_ReClassify();
				currentepoch = 0;
				AVQ_ResetVectors();
			}
		}

		if (verify)
			print_output(zold);

		if (report && (++rtime >= report)){
			rtime = 0;
#ifdef	DEBUG
			if (avq_flag)
				fprintf(stderr, "err:%f alpha:%f\n", avq_err, exp(-avq_phi * avq_err));
#endif
			if (ce == 2)
				print_error(&ce_err);
			else
				print_error(&err);
		}

		if (check && (++ctime >= check)){
			ctime = 0;
			save_wts();
		}

		if (++ttime >= tmax)
			ttime = 0;

		if (++utime >= umax){
			utime = 0;
			update_weights(wt,dwt,winc);
		}

		sweep++;
		err_sweeps++;
	}
	if (learning)
		save_wts();
	exit(0);

}

usage() {
  fprintf(stderr, "\n");
  fprintf(stderr, "-f fileroot:\tspecify fileroot <always required>\n");
  fprintf(stderr, "-l weightfile:\tload in weightfile\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "-s #:\trun for # sweeps <always required>\n");
  fprintf(stderr, "-q:\tperform AVQ on hidden unit activations to context units\n");
  fprintf(stderr, "-e #:\tAVQ: number of sweeps in an epoch <always required>\n");
  fprintf(stderr, "-p #:\tAVQ: phi parameter (default = 60)\n");
  fprintf(stderr, "-d #:\tAVQ: T parameter (default = 0.5)\n");
  fprintf(stderr, "-r #:\tset learning rate to # (between 0. and 1.) [0.1]\n");
  fprintf(stderr, "-m #:\tset momentum to # (between 0. and 1.) [0.0]\n");
  fprintf(stderr, "-n #:\t# of clock ticks per input vector [1]\n");
  fprintf(stderr, "-t:\tfeedback teacher values in place of outputs\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "-S #:\tseed for random number generator [random]\n");
  fprintf(stderr, "-U #:\tupdate weights every # sweeps [1]\n");
  fprintf(stderr, "-E #:\trecord rms error in .err file every # sweeps [0]\n");
  fprintf(stderr, "-C #:\tcheckpoint weights file every # sweeps [0]\n");
  fprintf(stderr, "-M #:\texit program when rms error is less than # [0.0]\n");
  fprintf(stderr, "-X:\tuse auxiliary .reset file\n");
  fprintf(stderr, "-P:\tprobe selected nodes on each sweep (no learning)\n");
  fprintf(stderr, "-V:\tverify outputs on each sweep (no learning)\n");
  fprintf(stderr, "-R:\tpresent input patterns in random order\n");
  fprintf(stderr, "-I:\tignore input values during extra clock ticks\n");
  fprintf(stderr, "-T:\tignore target values during extra clock ticks\n");
  fprintf(stderr, "-L:\tuse RTRL temporally recurrent learning\n");
  fprintf(stderr, "-B #:\toffset for offset biasi initialization\n");
  fprintf(stderr, "\n");
}

void
intr(sig)
	int	sig;
{
	save_wts();
	exit(sig);
}

